/*
 * Support for initialising program memory regions in new tasks.
 *
 * Copyright (C) 2022, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/sys/err.h>
#include <l4/util/elf.h>

#include <fsclient/client.h>
#include <systypes/fcntl.h>

#include <string.h>

#include "memory.h"



/* Obtain the payload as a dataspace for a named file. */

long exec_get_payload(const char *filename, Payload **payload, bool attach)
{
  file_t *file = client_open(filename, O_RDONLY);

  return exec_get_payload_file(file, payload, attach);
}

/* Obtain the payload as a dataspace for an existing file. */

long exec_get_payload_file(file_t *file, Payload **payload, bool attach)
{
  if ((file == NULL) || !client_opened(file))
    return -L4_EIO;

  /* Obtain metadata from the file. */

  char *buf = (char *) client_mmap(file, 0, file->size, 0, 0, L4RE_DS_F_R);

  /* Test the file type indicator. */

  if ((file->size < EI_NIDENT) || memcmp(buf, "\x7f" "ELF", 4))
    return -L4_EINVAL;

  /* Attempt to get a payload object appropriate for a particular object size
     variant. */

  *payload = exec_open_payload(buf);

  if (*payload == NULL)
    return -L4_ERANGE;

  if ((file->size < (*payload)->header_extent()) ||
      (file->size < (*payload)->program_header_extent()))
  {
    delete *payload;
    *payload = NULL;
    return -L4_EIO;
  }

  /* Obtain all loadable segments. */

  for (unsigned int i = 0; i < (*payload)->segments(); i++)
  {
    Segment *segment = (*payload)->segment(i);
    long err;

    if (!segment->loadable())
      continue;

    if (segment->file_contents())
    {
      file_t *rfile = client_reopen(file, file_opening_flags(segment->region_flags()));

      if ((rfile == NULL) || !client_opened(rfile))
      {
        delete *payload;
        *payload = NULL;
        return -L4_EIO;
      }

      err = segment->fill(rfile, attach);
    }
    else
      err = segment->allocate(attach);

    if (err)
    {
      delete *payload;
      *payload = NULL;
      return err;
    }
  }

  return L4_EOK;
}

/* vim: tabstop=2 expandtab shiftwidth=2
*/
